(* Read /dev/urandom.
 * Copyright (C) 2013 Red Hat Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *)

(** Read and return N bytes (only) from /dev/urandom. *)

val urandom_bytes : int -> string
(** Read N bytes from /dev/urandom and return it as a binary string. *)

val urandom_uniform : int -> string -> string
(** [urandom_uniform n chars] returns [n] bytes, uniformly
    distributed from the sets of characters [chars]. *)
